import logging
import pickle
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any, BinaryIO
from urllib.parse import urljoin

try:
    from tqdm.auto import tqdm

    TQDM_INSTALLED = True
except ModuleNotFoundError:
    TQDM_INSTALLED = False

_CHUNK_SIZE: int = 1024

LOGGER = logging.getLogger("rics.utility.misc")
_GLOR_LOGGER = LOGGER.getChild("get_local_or_remote")


def get_local_or_remote(
    file: str,
    *,
    remote_root: str,
    local_root: str,
    force: bool = False,
    postprocessor: Callable[[str], Any] | None = None,
    show_progress: bool | None = TQDM_INSTALLED,
) -> Path:
    """Retrieve a file from a local or remove (HTTP/HTTPS) source.

    Args:
        file: Name of an object to retrieve.
        remote_root: Remote location where the `file` may be downloaded.
        local_root: Local storage location for `file`.
        force: If ``True``, always fetch from the remote.
        postprocessor: A transformer that accepts a file path to a pickled object, and returns a transformed object.
            Returned value must be pickleable.
        show_progress: If ``True``, show a progress bar.

    Returns:
        Location where `file` may be accessed locally. Transformations are applied if a `postprocessor` is given, in
        which case the returned ``Path`` will point to a transformed result.
    """
    if show_progress is None:
        show_progress = TQDM_INSTALLED
    elif show_progress and not TQDM_INSTALLED:
        raise ModuleNotFoundError("The tqdm package is not installed; cannot show progress.")

    return _get_file(
        file=str(file),
        remote_root=str(remote_root),
        local_root=Path(str(local_root)).absolute(),
        force=force,
        postprocessor=postprocessor,
        show_progress=show_progress,
    )


def _get_file(
    file: str,
    remote_root: str,
    local_root: Path,
    force: bool,
    postprocessor: Callable[[str], Any] | None,
    show_progress: bool,
) -> Path:
    local_file_path = local_root.joinpath(file)
    remote_file_path = None if remote_root is None else urljoin(remote_root, file)

    _GLOR_LOGGER.debug(f"Local file path: '{local_file_path}'.")
    _GLOR_LOGGER.debug(f"Remote file path: '{remote_file_path}'.")

    if not local_root.is_dir():
        raise ValueError(f"Local root path '{local_root}' does not exist or is not a directory.")

    need_postprocessing = False
    if force or not local_file_path.exists():
        if remote_file_path is None:
            raise ValueError(f"Local file does not exist: '{local_file_path}'.")
        _fetch(local_file_path, remote_file_path, show_progress)
        if postprocessor is not None:
            need_postprocessing = True

    if postprocessor is not None:
        local_processed_file_path = local_root.joinpath(postprocessor.__name__).joinpath(file).with_suffix(".pkl")
        _GLOR_LOGGER.info(f"Local processed file path: '{local_processed_file_path}'.")
        need_postprocessing = need_postprocessing or not local_processed_file_path.exists()
        if force or need_postprocessing:
            local_processed_file_path.parent.mkdir(parents=True, exist_ok=True)
            _GLOR_LOGGER.info(f"Running {postprocessor.__name__}..")
            result = postprocessor(str(local_file_path))
            _GLOR_LOGGER.info(f"Serializing processed data to '{local_processed_file_path}'..")
            with local_processed_file_path.open("wb") as f:
                pickle.dump(result, f)

        return local_processed_file_path
    else:
        return local_file_path


def _fetch(local_file_path: Path, remote_file_path: str, show_progress: bool) -> None:
    try:
        import requests
    except ModuleNotFoundError:
        raise ModuleNotFoundError("The 'requests' package is not installed; cannot download remote content.") from None

    _GLOR_LOGGER.info(f"Fetching data from '{remote_file_path}'..")
    response = requests.get(remote_file_path, stream=True, timeout=300)
    chunks = response.iter_content(_CHUNK_SIZE)
    with local_file_path.open("wb") as output_file:
        if show_progress:
            total = response.headers.get("Content-Length")
            _with_progress(
                chunks,
                output_file,
                remote_file_path,
                total=None if not total else int(total),
            )
        else:
            _without_progress(chunks, output_file)


def _without_progress(chunks: Iterable[bytes], output_file: BinaryIO) -> None:
    for chunk in chunks:
        if chunk:
            output_file.write(chunk)


def _with_progress(
    chunks: Iterable[bytes],
    output_file: BinaryIO,
    remote_file_path: str,
    total: int | None,
) -> None:
    kwargs = {
        "total": total,
        "desc": str(remote_file_path),
        "unit": "iB",
        "unit_scale": True,
        "unit_divisor": _CHUNK_SIZE,
    }
    with tqdm(chunks, **kwargs) as progress:
        for chunk in chunks:
            if chunk:
                output_file.write(chunk)
                progress.update(_CHUNK_SIZE)
        if total is not None:
            progress.n = total



from collections.abc import Callable
from functools import update_wrapper
from typing import Any, Literal

import click

from rics.logs import DATE_FORMAT, FORMAT_SEC, LoggingSetupHelper, _UserVerbosityLevels

AnyCallable = Callable[..., Any]
Decorator = Callable[[AnyCallable], AnyCallable]
Mode = Literal["forward", "forward_both", "pop", "skip"]


class VerbosityParamType(click.types.IntParamType):
    name = "verbosity"

    def __init__(self, max: int) -> None:
        self._max = max

    def convert(self, value: Any, param: click.Parameter | None, ctx: click.Context | None) -> Any:
        inv_value: int = super().convert(value, param, ctx)
        if inv_value > self._max:
            self.fail(f"May be repeated at most {self._max} times (got {value}).", param, ctx)
        return inv_value


def logging_verbosity_option(
    *param_decls: str,
    mode: Mode = "pop",
    # Helper class params
    levels: _UserVerbosityLevels,
    format: str = FORMAT_SEC,
    datefmt: str = DATE_FORMAT,
    # Click params
    cls: type[click.Option] | None = None,
    **attrs: Any,
) -> Decorator:
    """Add a ``click`` option to a command.

    Mode options:
        * `forward`: Configure logging, then forward the parameter.
        * `forward_both`: Do **not** configure logging. Forward a tuple ``(verbosity, helper)`` instead.
        * `pop` (default): Configure logging and remove the parameter.
        * `skip`: Does **not** configure logging or create a helper. The parameter is forwarded.

    Args:
        *param_decls: Positional arguments to the constructor of ``cls``.
        mode: Logging setup mode. See above for options.
        levels: An iterable of levels, where each level is a dict ``{logger_name: log_level}``.
        format: Format string for emitted messages; see :func:`rics.logs.basic_config`.
        datefmt: Format string for date/time; see :func:`rics.logs.basic_config`.
        cls: Type of :class:`click.Option` to instantiate.
        **attrs: Passed as keyword arguments to the constructor of ``cls``.
            Defaults are provided for `help`, `count`, and `type`.

    Raises:
        TypeError: If `attrs` is incompatible with this method.

    See Also:
        The :func:`click.option` function and :meth:`rics.logs.LoggingSetupHelper` class.

    Examples:
        Decorating ``click`` commands.

        >>> import logging, click
        >>> @click.command
        >>> @logging_verbosity_option(
        ...     "--verbose", "-v",
        ...     levels=[
        ...         {"rics": "INFO", "id_translation": "WARNING"},
        ...         {"rics": "DEBUG"},
        ...     ],
        ... )
        >>> @click.pass_context
        >>> def cli(cxt: click.Context, verbose: int) -> None:
        >>>     print(verbose)
        >>>     print(logging.getLogger("rics"))
        >>>     print(logging.getLogger("id_translation"))

        When running this command, passing ``-vv`` increases `rics` verbosity to ``logging.DEBUG``.

        Advanced configuration with ``mode="forward_both"``.

        >>> import logging, click
        >>> from rics.logs import LoggingSetupHelper
        >>> @click.command
        >>> @logging_verbosity_option(
        ...     "--verbose", "-v",
        ...     mode="forward_both",
        ...     levels=[
        ...         {"rics": "INFO", "id_translation": "WARNING"},
        ...         {"rics": "DEBUG"},
        ...     ],
        ... )
        >>> @click.pass_context
        >>> def cli(
        ...     cxt: click.Context,
        ...     verbose: tuple[int, LoggingSetupHelper],
        ... ) -> None:
        >>>     verbosity, helper = verbose
        >>>     helper.configure_logging(verbosity)  # 0=logging.root.disabled=True
        >>>     if verbosity == 0:
        >>>         import os, sys
        >>>         sys.stdout = open(os.devnull, "w")
        >>>         cxt.meta["no_stdout"] = True

        In this mode, :meth:`.LoggingSetupHelper.configure_logging()` is not called before invoking ``cli()``. Above is
        a dummy program that disables ``sys.stdout`` if verbosity is zero (i.e. ``-v`` is repeated zero times).
    """
    helper = LoggingSetupHelper(levels, format=format, datefmt=datefmt)

    if "help" not in attrs:
        attrs["help"] = _create_help_string(helper.get_level_descriptions())

    if "count" in attrs:
        raise TypeError(f"{attrs['count']=} is not allowed.")
    attrs["count"] = True

    if "type" in attrs:
        raise TypeError(f"{attrs['type']=} is not allowed.")
    param_type = VerbosityParamType(helper.max)
    attrs["type"] = param_type

    click_option_decorator = click.option(*param_decls, cls=cls, **attrs)

    if mode == "skip":
        return click_option_decorator

    return _create_decorator(helper, click_option_decorator, param_type, mode)


def _create_decorator(
    helper: LoggingSetupHelper,
    click_decorator: Decorator,
    param_type: VerbosityParamType,
    mode: Mode,
) -> Decorator:
    def get_param_name(func: AnyCallable) -> str:
        assert hasattr(func, "__click_params__")  # noqa: S101
        cp: click.Parameter
        for cp in func.__click_params__:
            if cp.type is param_type:
                assert isinstance(cp.name, str)  # noqa: S101
                return cp.name

        raise RuntimeError(f"Unable to derive verbosity parameter name: {func.__click_params__=}.")

    def configure_logging_decorator(func: AnyCallable) -> AnyCallable:
        click_func = click_decorator(func)  # Sets __click_params__ on func.
        param_name = get_param_name(click_func)  # Could click_func be a click.Command at this point?

        def wrapper(*args: Any, **kwargs: Any) -> Any:
            verbosity = kwargs.pop(param_name) if mode == "pop" else kwargs[param_name]
            if not isinstance(verbosity, int):
                raise TypeError(f"kwargs[{param_name!r}={kwargs[param_name]!r}: expected `int`")

            if mode == "forward_both":
                kwargs[param_name] = verbosity, helper
                return func(*args, **kwargs)

            helper.configure_logging(verbosity)
            return func(*args, **kwargs)

        return update_wrapper(wrapper, click_func)

    return configure_logging_decorator


def _create_help_string(descriptions: list[str]) -> str:
    """Create help string."""
    lines = [
        "Controls application logging to stderr.",
        "\n",
        "\b\nVerbosity levels:",
        "  0 = Logging disabled (default).",
        *(f"  {verbosity} = {line}" for verbosity, line in enumerate(descriptions, start=1)),
        f"Repeat up to {len(descriptions)} times for increased verbosity.",
    ]
    return "\n".join(lines)



"""Utility functions that act on or produce strings."""

import typing as _t


def format_bytes(n: int, *, binary: bool = True, long: bool = False, decimals: int = 2) -> str:
    """Format bytes as a string.

    Args:
        n: Number of bytes. Must be positive.
        binary: Output `binary <https://en.wikipedia.org/wiki/Binary_prefix>`_ prefixes if ``True``,
            use `metric (SI) <https://en.wikipedia.org/wiki/Metric_prefix>`_ prefixes otherwise.
        long: Output out full unit and prefix if ``True``, use abbreviated versions otherwise.
        decimals: Number of decimals to include. Ignored for when `n < base`.

    Returns:
        Formatted number of bytes.

    Examples:
        **Formatting on prefix bounds**

        The jump as made at `base / 2`, where `base` is one of 1024 and 1000 (when ``binary=False``).

        >>> format_bytes(512 * 1024)
        '512.00 KiB'
        >>> format_bytes(512 * 1024 + 1)
        '0.50 MiB'

        This rule does *not* apply when `n <= base`.

        >>> format_bytes(1024, long=True)
        1024 bytes
        >>> format_bytes(1024 + 1)
        '1.00 KiB'

        **Output flags**

        >>> format_bytes(20190511, binary=False, long=False)
        '20.19 MB'
        >>> format_bytes(20190511, binary=False, long=True)
        '20.19 megabytes'
        >>> format_bytes(20190511, binary=True, long=False)
        '19.26 MiB'
        >>> format_bytes(20190511, binary=True, long=True)
        '19.26 mebibytes'

        **Large outputs**

        Metric and binary have different upper limits.

        >>> format_bytes(21**21, binary=True)
        '2416.44 YiB'
        >>> format_bytes(21**21, binary=True, long=True)
        '2416.44 yobibytes'
        >>> format_bytes(21**21, binary=False)
        '5.84 RB'
        >>> format_bytes(21**21, binary=False, long=True)
        '5.84 ronnabytes'

        If you ever see output like this, please let me know so that I can brag that someone important is using my
        little library.
    """
    base = 1024 if binary else 1000

    if n <= base:
        return f"{n} {'bytes' if long else 'B'}"

    x: float = n * 2.0
    n_divisions = -1
    while x > base:
        x /= base
        n_divisions += 1

    if binary:
        # https://en.wikipedia.org/wiki/Binary_prefix
        if long:
            prefixes = ["kibi", "mebi", "gibi", "tebi", "pebi", "exbi", "zebi", "yobi"]
        else:
            prefixes = ["Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"]
    else:  # noqa: PLR5501  # BUG: this rule doesn't preserve comments: https://github.com/astral-sh/ruff/issues/9790
        # https://en.wikipedia.org/wiki/Metric_prefix
        if long:
            prefixes = ["kilo", "mega", "giga", "tera", "peta", "exa", "zetta", "yotta", "ronna", "quetta"]
        else:
            prefixes = ["k", "M", "G", "T", "P", "E", "Z", "Y", "R", "Q"]

    try:
        prefix = prefixes[n_divisions]
    except IndexError:
        prefix = prefixes[-1]
        x = n / base ** len(prefixes)

    prefix += "bytes" if long else "B"
    return f"{x / 2:.{decimals}f} {prefix}"


def format_perf_counter(start: float, *, end: float | None = None, full: bool = False) -> str:
    """Format performance counter output.

    This function formats performance counter output based on the time elapsed. This is a thin wrapper around the
    :func:`~rics.strings.format_seconds` function.

    Args:
        start: Start time.
        end: End time. Retrieved using :py:func:`time.perf_counter` if ``None``.
        full: If ``True``, show all non-zero components above four hours.

    Returns:
        A formatted performance counter time.

    Examples:
        Basic usage.

        >>> import time
        >>> start = time.perf_counter()
        >>> time.sleep(1219.0)  # doctest: +SKIP
        >>> format_perf_counter(start)  # doctest: +SKIP
        '20m 19s'

        With no `end` argument given, the current time is retrieved using :py:func:`time.perf_counter`.
    """
    from time import perf_counter

    end = perf_counter() if end is None else end
    return format_seconds(end - start, full=full)


def format_seconds(t: float, *, allow_negative: bool = False, full: bool = False) -> str:
    """Format performance counter output.

    Args:
        t: Time in seconds.
        allow_negative: If ``True``, format negative `t` with a leading minus sign.
        full: If ``True``, show all non-zero components above four hours.

    Returns:
        A formatted performance counter time.

    Examples:
        Basic usage.

        >>> format_seconds(0.0000154)
        '15 μs'
        >>> format_seconds(0.154)
        '154 ms'
        >>> format_seconds(31.39)
        '31.4 sec'

        Clock units are used for `t > 60` seconds.

        >>> format_seconds(59.99)
        '60.0 sec'
        >>> format_seconds(60.00)
        '60.0 sec'
        >>> format_seconds(60.01)
        '1m'
        >>> format_seconds(309623.49)
        '3d 14h'

        Large intervals is rounded by default. You may set ``full=True`` to show full output.

        >>> format_seconds(309623.49)
        '3d 14h'
        >>> format_seconds(309633.51, full=True)
        '3d 14h 0m 34s'

    Raises:
        ValueError: If ``t < 0`` and ``allow_negative=False`` (the default).

    """
    if t < 0:
        if not allow_negative:
            allow_negative = True
            raise ValueError(f"Refuse to format {t=} < 0; to allow, set {allow_negative=}")
        return f"-{format_seconds(abs(t), full=full)}"

    long_limit: float = 60.0
    return _format_seconds(t) if t <= long_limit else _format_minutes(t, full)


def _format_minutes(t: float, full: bool) -> str:
    if full or t < 4 * 3600.0:
        total_seconds = round(t)
    else:
        total_seconds = 60 * round(t / 60)  # Drop seconds above four hours

    days, seconds = divmod(total_seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)

    parts = (days, hours, minutes, seconds)
    nonzero = tuple(p > 0 for p in parts)
    start = nonzero.index(True)
    stop = len(nonzero) - nonzero[::-1].index(True)
    return " ".join(f"{parts[i]}{'dhms'[i]}" for i in range(start, stop))


def _format_seconds(t: float) -> str:
    single_decimal_limit: float = 1.0
    if t >= single_decimal_limit:
        return f"{t:.1f} sec"

    double_decimal_limit: float = 0.5
    if t > double_decimal_limit:
        return f"{t:.2f} sec"

    if t > 10**-3:
        return f"{t * 10**3:.0f} ms"
    if t > 10**-6:  # 1 μs
        return f"{t * 10**6:.0f} μs"
    if t > 10**-9:
        return f"{t * 10**9:.0f} ns"

    return f"{t:.3g} sec"


def camel_to_snake(s: str) -> str:
    """Naive ``camelCase`` or ``PascalCase`` to ``snake_case`` conversion.

    Args:
        s: A string to convert.

    Returns:
        A ``snake_case`` string.

    Raises:
        IndexError: If `string` is empty.

    Examples:
        Converting camel case strings.

        >>> camel_to_snake("ClassName")
        'class_name'
        >>> camel_to_snake("variableName")
        'variable_name'

        Proper ``snake_case`` strings will not be changed.

        >>> camel_to_snake("already_snake_case")
        'already_snake_case'

    Notes:
        Passing ``SCREAMING_SNAKE_CASE`` strings is **not** supported.
    """
    parts = [s[0]]

    for ch in s[1:]:
        if ch.isupper():
            parts.append("_")
        parts.append(ch)

    return "".join(parts).lower()


def snake_to_camel(s: str, *, lower: bool = True) -> str:
    """Naive ``snake_case`` to ``camelCase`` conversion.

    Args:
        s: A string to convert.
        lower: If ``False``, return ``PamelCase`` instead of ``camelCase``.

    Returns:
        A ``camelCase`` string.

    Raises:
        IndexError: If `string` is empty.

    Examples:
        Converting snake case strings.

        >>> snake_to_camel("snake_case")
        'snakeCase'

        Passing ``SCREAMING_SNAKE_CASE`` strings is supported.

        >>> snake_to_camel("SCREAMING_SNAKE_CASE")
        'screamingSnakeCase'

        Set ``lower=False`` to convert to ``PascalCase`` or ``UpperCamelCase``.

        >>> snake_to_camel("SCREAMING_SNAKE_CASE", lower=False)
        'ScreamingSnakeCase'

    Notes:
        Passing ``camelCase`` strings is **not** supported.
    """
    s = s.title().replace("_", "")
    s0 = s[0]
    if lower:
        s0 = s0.lower()
    return s0 + s[1:]


TRUE = "1", "true", "yes", "on", "enable", "enabled"
FALSE = "0", "false", "no", "off", "disable", "disabled"


def str_as_bool(s: str) -> bool:
    """Convert a string `s` to a boolean value.

    The output is determined by the content of `s`, as per the mapping shown below.

    Keys:
        * False: ``{false}``
        * True: ``{true}``

    Matching is case-insensitive.

    Args:
        s: A string.

    Returns:
        A ``bool`` value.

    Raises:
        TypeError: If `s` is not a string.
        ValueError: If `s` cannot be converted to ``bool`` using the keys above.

    Examples:
        Basic usage.

        >>> str_as_bool("true"), str_as_bool("false")
        (True, False)

        The input is cleaned and normalized.

        >>> str_as_bool(" TRUE"), str_as_bool("False")
        (True, False)

        Input strings are normalized using :py:meth:`str.strip` and :py:meth:`str.lower`.

    Notes:
        Using ``bool(<str>)`` is equivalent to ``len(<str>) == 0``.
    """
    if not isinstance(s, str):
        msg = f"Input must be a string; got {type(s).__name__}."
        raise TypeError(msg)

    s = s.strip().lower()
    if s in FALSE:
        return False
    if s in TRUE:
        return True

    error = ValueError(f"Cannot cast {s!r} to `bool`.")
    error.add_note(f"{FALSE=}")
    error.add_note(f"{TRUE=}")
    raise error


if str_as_bool.__doc__:
    str_as_bool.__doc__ = str_as_bool.__doc__.format(false=FALSE, true=TRUE)


def format_kwargs(
    kwargs: _t.Mapping[str, _t.Any],
    *,
    max_value_length: int = 120,
    prefix_classname: bool = False,
    include_module: bool = False,
) -> str:
    """Format keyword arguments.

    Args:
        kwargs: Arguments to format.
        prefix_classname: If ``True``, prepend the class name if a value belongs to a class.
        include_module: If ``True``, prepend the public module (see :func:`.misc.get_public_module`).
        max_value_length: Replace value with the class name above this limit. 0=no limit.

    Returns:
        A string on the form `'key0=repr(value0), key1=repr(value1)'`.

    Raises:
        ValueError: For keys in `kwargs` that are not valid Python argument names.

    Examples:
        Basic usage.

        >>> format_kwargs({"an_int": 1, "a_string": "Hello!"})
        "an_int=1, a_string='Hello!'"

    Notes:
        Uses :class:`ReprFormatter` to format values.
    """
    invalid = [k for k in kwargs if not k.isidentifier()]
    if invalid:
        raise ValueError(f"Got {len(invalid)} invalid identifiers: {invalid}.")

    rf = ReprFormatter(
        max_value_length=max_value_length,
        prefix_classname=prefix_classname,
        include_module=include_module,
    )

    return ", ".join(f"{k}={rf.format_value(v)}" for k, v in kwargs.items())


class ReprFormatter:
    """Alternative :py:func:`repr` implementation.

    Values above `max_value_length` characters are replaced by stylized class names.

    Args:
        max_value_length: Use class name above this length. 0=no limit, -1=force class name.
        prefix_classname: If ``True``, prepend the class name if a value belongs to a class.
        include_module: If ``True``, prepend the public module (see :func:`.misc.get_public_module`).
        module_aliases: A mapping of module replacements, e.g. ``{"pandas": "pd"}``. Default is
            :attr:`DEFAULT_MODULE_ALIASES`. Trailing dots are added automatically. Ignored when `include_module` is
            ``False``.

    See Also:
        The :func:`format_kwargs`, :func:`.misc.tname`, and :func:`.misc.get_public_module` functions.
    """

    DEFAULT_MODULE_ALIASES: _t.Mapping[str, str] = {
        "numpy": "np",
        "pandas": "pd",
        "polars": "pl",
        "tensorflow": "tf",
        "matplotlib.pyplot": "plt",
    }

    def __init__(
        self,
        *,
        max_value_length: int = 120,
        prefix_classname: bool = False,
        include_module: bool = False,
        module_aliases: _t.Mapping[str, str] | None = None,
    ) -> None:
        self._max_value_length = max_value_length

        if module_aliases is None:
            module_aliases = self.DEFAULT_MODULE_ALIASES
        self._module_aliases = {k + ".": v + "." for k, v in module_aliases.items()}

        self._prefix_classname = prefix_classname
        self._include_module = include_module

        self._cache: dict[int, str] = {}

    def format_value(self, value: _t.Any) -> str:
        """Convert any value to string."""
        value_id = id(value)
        value_repr = self._cache.get(value_id)
        if value_repr is None:
            value_repr = self._format_value(value)
            self._cache[value_id] = value_repr

        return value_repr

    def _format_value(self, value: _t.Any) -> str:
        """Convert any value to string."""
        if self._max_value_length == 0:
            return self._serialize_as_value(value)
        if self._max_value_length < 0:
            shape = self._get_shape(value)
            return self._serialize_as_class(value, shape)

        for serializer in [
            self._repr_str,
            self._repr_builtin_collection,
            self._format_ndim_array,
        ]:
            value_repr = serializer(value)
            if isinstance(value_repr, str):
                return value_repr
            elif value_repr is False:
                break

        value_repr = self._serialize_as_value(value)
        if len(value_repr) <= self._max_value_length:
            return value_repr

        return self._serialize_as_class(value, ())

    def format_ndim_array(self, value: _t.Any) -> str:
        """Format shaped types, e.g. attr:`pandas.DataFrame.shape`."""
        shape = self._get_shape(value)
        if shape:
            return self._serialize_as_class(value, shape)

        msg = f"{type(value).__name__}.shape={shape} not valid"
        raise TypeError(msg)

    def _format_ndim_array(self, value: _t.Any) -> str | None:
        if shape := self._get_shape(value):
            return self._serialize_as_class(value, shape)
        return None

    @classmethod
    def _serialize_as_value(cls, value: _t.Any) -> str:
        from pprint import PrettyPrinter

        pp = PrettyPrinter(
            indent=2,
            width=120,
            depth=4,
            compact=True,
            sort_dicts=True,
            underscore_numbers=True,
        )
        return pp.pformat(value)

    def _serialize_as_class(self, value: _t.Any, shape: tuple[int, ...]) -> str:
        from rics.misc import tname

        value_cls = tname(value, prefix_classname=self._prefix_classname, include_module=self._include_module)

        if self._include_module:
            for module, alias in self._module_aliases.items():
                if value_cls.startswith(module):
                    value_cls = value_cls.replace(module, alias)

        if not shape:
            return value_cls

        dims = "x".join(map(str, shape))
        return f"{value_cls}[{dims}]"

    def _repr_str(self, value: _t.Any) -> str | bool:
        if isinstance(value, str):
            sz = len(value)
            if sz > self._max_value_length:
                return f"str[{sz}]"
            else:
                return repr(value)  # Might be longer than max_value_length if there's a lot of escaping.

        return True

    def _repr_builtin_collection(self, value: _t.Any) -> str | bool:
        if isinstance(value, (list, tuple, set)):
            if len(value) * 3 > self._max_value_length:
                shape = (len(value),)
                return self._serialize_as_class(value, shape)
            else:
                return False

        if isinstance(value, dict):
            if len(value) * 6 > self._max_value_length:
                shape = (len(value),)
                return self._serialize_as_class(value, shape)
            else:
                return False

        return True

    @classmethod
    def _get_shape(cls, value: _t.Any) -> tuple[int, ...]:
        if hasattr(value, "shape") and isinstance(value.shape, tuple):
            return value.shape
        elif hasattr(value, "__len__"):
            return (len(value),)
        return ()



"""Multivariate performance testing from the command line.

For details, see the :doc:`CLI documentation </documentation/cli/notebooks/timeit>` page.
"""

import datetime
import importlib
import inspect
import os
import shutil as _shutil
import sys
import warnings as _warnings
from pathlib import Path as _Path
from typing import Any as _Any
from typing import Never as _Never

import click
import pandas as pd

from rics import click as _rclick

from .._just_the_way_i_like_it import configure_stuff as _configure_stuff
from ._multi_case_timer import MultiCaseTimer as _Timer
from ._util import get_best as _get_best
from ._wrapper import run_multivariate_test as _run

LARGE_RESULT_ROW_LIMIT: int = 100_000
"""Ask for confirmation before writing results frames above this length."""

_show_warning_orig = _warnings.showwarning


@click.command(
    "timeit",
    epilog="""
    \b
    The CLI does not support facets (rows and columns). Please use
        https://rics.readthedocs.io/en/stable/api/rics.performance.html
    if your test data should be categorized into subgroups.
    """,
)
@click.pass_context
@click.option(
    "--time-per-candidate",
    "-t",
    default=2.0,
    help="Time in seconds to allocate to each candidate function.",
    type=float,
    show_default=True,
)
@click.option(
    "--name",
    "-n",
    default="performance.png",
    help="Name to use for artifacts produced. Also used as the figure title (stylized).",
    type=click.Path(dir_okay=False, writable=True),
    show_default=True,
)
@click.option(
    "--create/--no-create",
    default=False,
    help="Create files 'candidates.py' and 'test_data.py' and run a demo. Will not overwrite existing files.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--per-candidate/--no-per-candidate",
    default=True,
    help="Enable to print per-candidate best times. Just shows the best overall per data if disabled.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--yes",
    "-y",
    help=f"Automatic yes to prompts for things like writing large (>{LARGE_RESULT_ROW_LIMIT:_d} rows) result files.",
    is_flag=True,
)
@click.option(
    "--plot/--no-plot",
    help="Plotting requires `seaborn`.",
    is_flag=True,
)
def main(
    cxt: click.Context,
    time_per_candidate: float,
    name: str,
    create: bool,
    per_candidate: bool,
    yes: bool,
    plot: bool,
) -> None:
    """Multivariate performance testing.

    This is the https://pypi.org/project/rics/ version of the python timeit module. It may be used to run performance
    tests evaluating one or more candidate functions ('candidates.py') on one or more different kinds of inputs
    ('test_data.py').

    \b
    This script will:
        0. Create 'candidates.py' and 'test_data.py' (only if --create is set; will never overwrite).
        1. Quickly evaluate each candidate on all test data "a few times".
        2. Decide how many times to evaluate each candidate, such that the
           --time-per-candidate argument is respected.
        3. Print the best times per candidate/test_data
           combination to stdout.
        4. Save a performance overview figure to disk.
        5. Save raw timing data to disk as CSV.

    \b
    Required files:
        candidates.py - Members starting with 'candidate_' are used as candidates.
        test_data.py - Members starting with 'case_' are used as the case case data.

    Hint: Define a 'ALL'-attributes in 'candidates' and 'test_data' to declare explicit members to use.
    """  # noqa: D301
    _deprecated(cxt)

    if create:
        name, time_per_candidate = _create_input_files()

    name_path = _Path(name).with_suffix(".jpg").absolute()
    title = name_path.stem.replace("-", " ").replace("_", " ").title()

    width = _width()
    c6 = f"^{_width() - 6}"
    click.secho(f"{' Begin Performance Evaluation ':=^{width}}", fg="green")
    click.secho(f"|  {f'  {title!r}  ':{c6}}  |", fg="green")
    click.secho("-" * width, fg="green")

    # force=False has no effect during normal operation (i.e. when running from CLI), but keeps things tidy for CI/CD.
    os.environ["JTWILI"] = "true"  # Suppress warnings

    _configure_stuff(
        logging=cxt.meta.get("verbosity") is None,  # Configure logging only when invoked directly.
        ghost=False,
    )

    try:
        sys.path.insert(0, str(_Path().absolute()))
        candidates = _get_candidates()
        test_data = _get_test_data()
    finally:
        sys.path.pop(0)

    now = datetime.datetime.now()
    eta_str = (now + datetime.timedelta(seconds=len(candidates) * time_per_candidate)).strftime("%A %d, %H:%M:%S")
    now_str = now.strftime("%Y-%m-%d %H:%M:%S")

    w2 = f"<{_width() - 2}"
    click.secho(f"{f'| Found {len(candidates)} candidates and {len(test_data)} data variants:':{w2}} |", fg="green")
    click.secho(f"{f'|  -    Candidates: {[*_Timer.process_candidates(candidates)]}':{w2}} |", fg="green")
    click.secho(f"{f'|  - Data variants: {[*_Timer.process_test_data(test_data)]}':{w2}} |", fg="green")
    click.secho(f"{f'| Started: {now_str}, ETA: {eta_str}':{w2}} |", fg="green")
    click.secho("=" * width, fg="green")

    try:
        _warnings.showwarning = _show_warnings

        result = _run(candidates, test_data, time_per_candidate=time_per_candidate, plot=plot, show=False)
    except ValueError as e:  # pragma: no cover
        _handle_value_error(e)
        raise
    finally:
        _warnings.showwarning = _show_warning_orig

    _print_best_result(per_candidate, result, title)
    _save_plot(name_path, title)
    _save_report(name_path, result, not yes)


def _deprecated(cxt: click.Context) -> None:
    name = cxt.info_name
    if name != "mtimeit":
        return

    user_command = _rclick.get_user_command().replace(name, "rics timeit", 1)
    msg = (
        f"WARNING: Entrypoint `{name}` is deprecated."
        f"\nUse `{user_command}` instead, or pin `rics==5.0.1` to hide this warning."
    )
    click.secho(msg, err=True, fg="yellow")


def _get_test_data() -> _Any:
    try:
        import test_data as test_data_module  # type: ignore

        if hasattr(test_data_module, "ALL"):
            test_data = test_data_module.ALL
        else:
            test_data = {
                name.removeprefix("case_").replace("_", " "): data
                for name, data in inspect.getmembers(test_data_module)
                if name.lower().startswith("case_")
            }
    except ModuleNotFoundError:
        _missing_file("test_data.py")
    return test_data


def _get_candidates() -> _Any:
    try:
        candidates_module = importlib.import_module("candidates")

        if hasattr(candidates_module, "ALL"):
            candidates = candidates_module.ALL
        else:
            candidates = [func for name, func in inspect.getmembers(candidates_module) if name.startswith("candidate_")]
    except ModuleNotFoundError:
        _missing_file("candidates.py")
    return candidates


def _missing_file(name: str) -> _Never:
    cxt = click.get_current_context()
    click.secho(
        f"No file called '{name}' in the current working directory."
        f"\nHint: Create by running `{cxt.command_path} --create`.",
        fg="red",
        err=True,
    )
    sys.exit(2)


def _print_best_result(per_candidate: bool, result: pd.DataFrame, title: str) -> None:
    width = _width()
    c6 = f"^{_width() - 6}"
    click.secho("=" * width, fg="green")
    click.secho(f"|  {'  Best Times  ':{c6}}  |", fg="green")
    click.secho(f"|  {f'  {title!r}  ':{c6}}  |", fg="green")
    click.secho("=" * width, fg="green")
    click.secho(_get_best(result, per_candidate=per_candidate), fg="red")
    click.secho("=" * width, fg="green")


def _save_report(name_path: _Path, result: pd.DataFrame, prompt: bool) -> None:
    performance_report_path = name_path.with_suffix(".csv")

    if prompt and len(result) > LARGE_RESULT_ROW_LIMIT:
        click.secho(
            f"WARNING: The full timing report has {len(result):_d} rows, which may take a while to serialize.",
            fg="yellow",
        )
        if not click.confirm(f"Really print full report to '{performance_report_path}'?"):
            sys.exit(0)

    result.set_index(["Candidate", "Test data"]).to_csv(performance_report_path)
    click.secho(f"Data saved: '{performance_report_path}'", fg="green")


def _save_plot(name_path: _Path, title: str) -> None:
    import matplotlib.pyplot as plt

    fig = plt.gcf()
    fig.suptitle(title)
    plt.savefig(name_path)
    click.secho(f"Figure saved: '{name_path}'", fg="green")


def _create_input_files() -> tuple[str, float]:
    time_per_candidate = 0.1
    name = "create-example-run"

    for template_name in "candidates", "test_data":
        path = _Path().joinpath(template_name).with_suffix(".py").absolute()
        if path.exists():
            click.secho(
                f"ABORT: Refusing to overwrite existing file '{path}'. "
                f"Delete files 'candidates.py' and 'test_data.py' in '{path.parent}' to continue.",
                fg="red",
                err=True,
            )
            sys.exit(1)
        text = _Path(__file__).parent.joinpath(f"templates/{template_name}.py.txt").read_text()
        path.write_text(text)

    return name, time_per_candidate


def _handle_value_error(e: ValueError) -> None:
    msg = str(e)

    file: str | None = None
    if msg.startswith("No candidates"):
        msg += " Candidate functions must start with 'candidate_'."
        file = "candidates.py"
    elif msg.startswith("No case"):
        msg += " Case data attributes must start with 'case_'."
        file = "candidates.py"

    if file is None:
        return

    msg += f" Alternatively, you may define an attribute  'ALL = {{label: ...}}' in '{file}'."
    click.secho("ABORT: " + msg, fg="red")
    sys.exit(-1)


def _width() -> int:
    width = _shutil.get_terminal_size()[0]

    return min(width, 120)


def _show_warnings(message: Warning | str, category: type[Warning], *args: _Any, **kwargs: _Any) -> None:
    if isinstance(message, UserWarning) and message.args[0].startswith("The test results may be unreliable"):
        click.secho(f"WARNING: {message}", fg="yellow")
        return
    _show_warning_orig(message, category, *args, **kwargs)


if __name__ == "__main__":
    main()



import sys

import click

USER_COMMAND_KEY: str = "USER_COMMAND"


def get_user_command(cxt: click.Context | None = None, *, meta: bool = True) -> str:
    """Recreate the user CLI command.

    Args:
        cxt: Click context. Derive is ``None``.
        meta: If ``True``, use :attr:`click.Context.meta` caching.

    Returns:
        A user command such as ``python -m rics.cli -v kernel --name=my-kernel``.
    """
    if cxt is None:
        cxt = click.get_current_context(silent=False).find_root()

    if not meta:
        return _build_user_command(cxt)

    if user_command := cxt.meta.get(USER_COMMAND_KEY):
        assert isinstance(user_command, str)  # noqa: S101
        return user_command

    user_command = _build_user_command(cxt)
    cxt.meta[USER_COMMAND_KEY] = user_command
    return user_command


def _build_user_command(cxt: click.Context) -> str:
    return " ".join([cxt.command_path, *sys.argv[1:]])



from typing import overload

from ._base import read_env


@overload
def read_str(var: str, default: str = "", *, strict: bool = True, split: str) -> list[str]: ...
@overload
def read_str(var: str, default: str = "", *, strict: bool = True, split: None = None) -> str: ...
def read_str(var: str, default: str = "", *, strict: bool = True, split: str | None = None) -> str | list[str]:
    """Read ``str`` variable.

    Args:
        var: Variable to read.
        default: Default value to use when `var` is not set (or blank).
        strict: If ``False``, always fall back to `default` instead of raising.
        split: Character to split on. Returns ``list[str]`` when set.

    Returns:
        An ``str`` value, or a list thereof (if `split` is set).

    Examples:
        Basic usage.

        >>> import os
        >>> os.environ["MY_STR"] = "  foo "
        >>> read_str("MY_STR")
        'foo'

        Input is stripped and cleaned, returning bare strings. Empty items are discarded when using `split`.

        >>> os.environ["MY_STR_LIST"] = "  foo, , bar "
        >>> read_str("MY_STR_LIST", split=",")
        ['foo', 'bar']

        This means that input such as ``',  , , '`` will return an empty list.
    """
    return read_env(var, str.strip, default, strict=strict, split=split)



from pathlib import Path

from rics.paths import AnyPath

from ._patch_notes import PatchNotes
from ._write_changelog import write_changelog


def split_changelog(output_dir: AnyPath, changelog: AnyPath = "CHANGELOG.md") -> None:
    """Split a Markdown changelog into multiple pages.

    Args:
        output_dir: Output directory where pages will be placed. Only ``index.rst`` needs to be manually imported.
        changelog: The changelog to split.

    """
    with Path(changelog).open() as f:
        notes, refs = PatchNotes.from_markdown(f.read())

    write_changelog(output_dir, notes, refs)



"""Types used by this package."""

import os as _os
import typing as _t
from collections import abc as _abc
from enum import Enum as _Enum
from pathlib import Path as _Path

AnyPath: _t.TypeAlias = str | _os.PathLike[str] | _Path
"""Any path-like type; see :func:`~rics.paths.parse_any_path`."""

T = _t.TypeVar("T")
"""A type var with no type bounds."""
EnumT = _t.TypeVar("EnumT", bound=_Enum)
"""A type var bounded by :py:class:`enum.Enum`."""


def verify_enum(
    value: str | EnumT,
    enum_type: type[EnumT],
    *,
    name: str | None = None,
) -> EnumT:
    """Verify enum value.

    Args:
        value: User input value; either a ``str`` or an enum.
        enum_type: Desired enum type.
        name: Name of the user-facing argument. Used in error messages. Derive if ``None``.

    Returns:
        A enum of the given `enum_type`.

    Examples:
        Basic usage.

        >>> from enum import Enum, auto
        >>> class Response(Enum):
        ...     Yes = auto()
        ...     No = auto()

        Matching is case-insensitive when strings are given.

        >>> verify_enum("NO", Response)
        '<Response.No: 2>'

        Enums may also be passed.

        >>> verify_enum(Response.Yes, Response)
        '<Response.Yes: 1>'

        The default variable name used for errors is derived using :func:`rics.strings.camel_to_snake`.

        >>> verify_enum("x", Response)  # doctest: +SKIP
        TypeError: Bad response='x'; expected a Response enum option: Response.Yes | Response.No).

    Notes:
        This function wraps :meth:`LiteralHelper.from_enum`.
    """
    return LiteralHelper.from_enum(enum_type, default_name=name).check(value)


def verify_literal(
    value: _t.Any,
    literal_type: _abc.Collection[T] | type[T] | _t.Any,  # Runtime type is typically <typing special form>.
    *,
    name: str = "value",
    type_name: str | None = None,
    exc_type: type[Exception] | None = None,
    normalizer: _abc.Callable[[_t.Any], _t.Any] | None = None,
) -> T:
    """Verify enum value.

    Args:
        value: User input value; either a ``str`` or an enum.
        literal_type: A ``typing.Literal`` type alias or a collection of permitted values.
        name: Name of the user-facing argument. Used in error messages.
        type_name: Name of the type itself. Used in error messages.
        exc_type: Exception type to raise. Default is :py:class:`TypeError`.
        normalizer: A callable ``(value | option) -> new_value`` to use if `value` does not match exactly.

    Returns:
        A valid ``Literal`` value.

    Examples:
        Basic usage.

        >>> from typing import Literal
        >>> verify_literal("no", Literal["yes", "no"])
        'no'

        Note that the output type will be ``Any`` unless a collection is used (a named ``TypeAlias`` is not enough).
        This is due to a limitation in the Python typing machinery. To get around this, you need to use an explicitly
        typed output variable.

        >>> YesOrNot = Literal["yes", "no"]
        >>> yes_or_no: YesOrNot = verify_literal("Yes!", YesOrNot)  # doctest: +SKIP
        TypeError: Bad value='Yes!'; expected one of ('yes', 'no').

        By default, a :py:class:`TypeError` is raised if the input value does not match the literal type.

        .. hint::

           See the :class:`LiteralHelper` class docs for more examples.
    """
    return LiteralHelper[T](
        literal_type=literal_type,
        default_name=name,
        type_name=type_name,
        exc_type=exc_type,
        normalizer=normalizer,
    ).check(value)


class LiteralHelper(_t.Generic[T]):
    """Support class for :func:`verify_literal`.

    Using this class may improve performance when the same literal type is verified multiple times.

    Examples:
        Basic usage.

        >>> from typing import Literal
        >>> YesOrNo = Literal["yes", "no"]
        >>> helper = LiteralHelper[YesOrNo](YesOrNo, type_name="YesOrNo")

        The original value is returned as-is.

        >>> helper.check("yes")
        'yes'

        The helper class itself is callable.

        >>> helper("YES")  # doctest: +SKIP
        TypeError: Bad value='YES'; expected a YesOrNo['yes', 'no'].

        Pass an explicit `name` to customize the error message for a single check call.

        >>> helper("YES", name="user input")  # doctest: +SKIP
        TypeError: Bad user input='YES'; expected a YesOrNo['yes', 'no'].

        Using a normalizer.

        >>> normalizing_helper = LiteralHelper(YesOrNo, normalizer=str.lower)
        >>> normalizing_helper("YES")
        'yes'

        The normalizer is applied only if the original value doesn't match exactly.

    Notes:
        Explicitly typed helpers (e.g. ``LiteralHelper[YesOrNo](...)``) don't require typed output variables, even when
        using type aliases instead of collections. This is another advantage of using the helper instead of
        :func:`verify_literal`.
    """

    def __init__(
        self,
        literal_type: _abc.Collection[T] | type[T] | _t.Any,  # Runtime type is typically <typing special form>.
        *,
        default_name: str = "value",
        type_name: str | None = None,
        exc_type: type[BaseException] | None = None,
        normalizer: _abc.Callable[[_t.Any | T], _t.Any] | None = None,
    ) -> None:
        options = self._extract_options(literal_type)
        if not options:
            raise ValueError(f"Could not derive options from {literal_type!r}.")

        self._name = default_name
        self._type_name = type_name
        self._exc_type: type[BaseException] = TypeError if exc_type is None else exc_type
        self._options = options
        self._normalizer = normalizer

    @classmethod
    def from_enum(cls, enum_type: type[EnumT], *, default_name: str | None = None) -> "LiteralHelper[EnumT]":
        """Construct helper for an :class:`enum.Enum` type.

        Args:
            enum_type: Desired enum type.
            default_name: Default name of the user-facing argument. Used in error messages. Derive if ``None``.

        Returns:
            A ``LiteralHelper`` constructor.
        """
        if default_name is None:
            from rics.strings import camel_to_snake

            default_name = camel_to_snake(enum_type.__name__)

        return LiteralHelper(
            literal_type=enum_type,
            default_name=default_name,
            type_name=None,  # Derived and pretty-printed on failure if None.
            normalizer=cls._enum_normalizer,
        )

    @_t.overload
    def read_env(self, variable: str, *, default: T | None, split: str) -> list[T]: ...
    @_t.overload
    def read_env(self, variable: str, *, default: T | None, split: None = None) -> T: ...
    def read_env(self, variable: str, *, default: T | None, split: str | None = None) -> T | list[T]:
        """Read environment value.

        Args:
            variable: Environment variable name.
            default: Value to use is `variable` is not set or blank. Pass ``None`` to raise.
            split: Character to split on. Returns ``list[T]`` when set.

        Returns:
            A :attr:`T`, or a list thereof (if `split` is set).

        Notes:
            Use the :mod:`rics.env.read` functions to read primitive types such as booleans.
        """
        import os

        value = os.environ.get(variable, "").strip()
        if default is None:
            if variable not in os.environ:
                self._raise("<not set>", variable)
            elif not value:
                self._raise("<blank>", variable)
        elif not value:
            return default if split is None else [default]

        if split is None:
            return self.check(value, name=variable)

        values = value.split(split)
        return [self.check(value, name=f"{variable}[{i}]") for i, value in enumerate(map(str.strip, values)) if value]

    @classmethod
    def _enum_normalizer(cls, arg: _t.Any) -> str:
        return arg.name.lower() if isinstance(arg, _Enum) else str(arg).strip().lower()

    @property
    def options(self) -> tuple[T, ...]:
        """Permitted options as explicit values."""
        return self._options

    def check(self, value: _t.Any, name: str | None = None) -> T:
        """Alias of ``__call__``.

        Args:
            value: User input value.
            name: Name to use for this call.

        Returns:
            A valid value.
        """
        name = self._name if name is None else name
        options = self._options

        if value in options:
            return value  # type: ignore[no-any-return]

        normalizer = self._normalizer
        if normalizer:
            normalized_value = normalizer(value)
            for option in options:
                if normalizer(option) == normalized_value:
                    return option

        self._raise(repr(value), name)

    __call__ = check

    def _raise(self, value: str, name: str) -> _t.Never:
        options = self._options

        if self._type_name is None:
            if isinstance(options[0], _Enum):
                pretty_options = f"a {type(options[0]).__name__} enum option: {{ {' | '.join(map(str, options))} }}"
            else:
                pretty_options = f"one of {options}"
        else:
            pretty_options = f"a {self._type_name}[{', '.join(map(repr, options))}]"

        msg = f"Bad {name}={value}; expected {pretty_options}."
        raise self._exc_type(msg)

    @classmethod
    def _extract_options(cls, literal_type: _abc.Collection[T] | _t.Any) -> tuple[T, ...]:
        if isinstance(literal_type, type(_t.Literal[None])):
            return (*_t.get_args(literal_type),)

        if isinstance(literal_type, _abc.Collection) and not isinstance(literal_type, str):
            return (*literal_type,)

        union_type = type(_t.Union[str, int])  # noqa: UP007
        if isinstance(literal_type, union_type):
            from_union = cls._from_union(literal_type)
            if from_union:
                return from_union

        msg = f"Invalid type: {literal_type!r}. Expected a Literal, Union of Literal, Collection, or Enum."
        raise TypeError(msg)

    @classmethod
    def _from_union(cls, literal_type: _t.Any) -> tuple[T, ...]:
        literal_cls = type(_t.Literal[None])

        literals = []
        types = _t.get_args(literal_type)
        for t in types:
            if not isinstance(t, literal_cls):
                return ()
            for literal in _t.get_args(t):
                if literal not in literals:
                    literals.append(literal)

        return (*literals,)



import os
from collections.abc import Callable
from typing import NamedTuple, overload

from rics.types import T


@overload
def read_env(
    var: str,
    converter: Callable[[str], T],
    default: T,
    *,
    strict: bool = True,
    type_name: str | None = None,
    split: None = None,
    catch: tuple[type[Exception], ...] | None = None,
) -> T: ...


@overload
def read_env(
    var: str,
    converter: Callable[[str], T],
    default: T,
    *,
    strict: bool = True,
    type_name: str | None = None,
    split: str,
    catch: tuple[type[Exception], ...] | None = None,
) -> list[T]: ...


def read_env(
    var: str,
    converter: Callable[[str], T],
    default: T,
    *,
    strict: bool = True,
    type_name: str | None = None,
    split: str | None = None,
    catch: tuple[type[Exception], ...] | None = None,
) -> T | list[T]:
    """Read and convert an environment variable.

    Args:
        var: Variable to read.
        converter: A callable ``(str) -> T``, where the argument is the environment variable value.
        default: Default value to use when `var` is not set (or blank).
        strict: If ``False``, always fall back to `default` instead of raising.
        type_name: Used in error messages. Derive if ``None``.
        split: Character to split on. Returns ``list[T]`` when set.
        catch: Types to suppress when calling ``converter``. Default is ``(ValueError, TypeError)``.

    Returns:
        A ``T`` value, or a list thereof (if `split` is set).

    Raises:
        ValueError: If conversion fails.

    Notes:
        If the `variable` key is unset or the value is empty, the `default` value is always returned.
    """
    value = os.environ.get(var)
    if value is None:
        return [default] if split else default

    value = str(value).strip()  # Just in case; should already be a string.
    if value == "":
        return [default] if split else default

    if catch is None:
        catch = ValueError, TypeError

    cause: BaseException
    reason: str
    if split is None:
        try:
            return converter(value)
        except catch as exc:
            cause = exc
            reason = f": {exc}"
    else:
        result = _split(value.split(split), converter, catch)
        if isinstance(result, _ExceptionDetails):
            reason = f".\nNOTE: Failed at {var}[{result.idx}]={result.value!r}: {result.exception}"
            cause = result.exception
        else:
            return result

    if strict:
        if type_name is None:
            type_name = type(default).__name__
        if split:
            type_name = f"list[{type_name}]"
        msg = f"Bad value {var}={value!r}; not a valid `{type_name}` value" + reason
        raise ValueError(msg) from cause

    return [default] if split else default


class _ExceptionDetails(NamedTuple):
    exception: BaseException
    idx: int
    value: str


def _split(
    values: list[str],
    converter: Callable[[str], T],
    catch: tuple[type[BaseException], ...],
) -> list[T] | _ExceptionDetails:
    items = []
    for i, value in enumerate(values):
        stripped = value.strip()
        if stripped:
            try:
                result = converter(stripped)
                items.append(result)
            except catch as exc:
                return _ExceptionDetails(exc, i, stripped)

    return items



from collections.abc import Collection, Iterable
from typing import Any

import pandas as pd

from ._multi_case_timer import CandidateMethodArg, MultiCaseTimer, TestDataArg
from ._util import to_dataframe
from .types import DataFunc, DataType, Ts


def run_multivariate_test(
    candidate_method: CandidateMethodArg[DataType],
    test_data: TestDataArg[DataType] | DataFunc[*Ts, DataType],  # DataFunc[DataFuncP, DataType]
    *,
    # case_args: Collection[tuple[*Ts]] | None = None,
    # kwargs: Any | None = None,
    time_per_candidate: float = 6.0,
    plot: bool = True,
    show: bool = True,
    names: Iterable[str] | None = (),
    progress: bool = False,
    case_args: Collection[tuple[*Ts]] | None = None,
    kwargs: Any | None = None,
    **plot_kwargs: Any,
) -> pd.DataFrame:
    """Run performance tests for multiple candidate methods on collections of test data.

    This is a convenience method which combines :meth:`MultiCaseTimer.run() <rics.performance.MultiCaseTimer.run>`,
    :meth:`~rics.performance.to_dataframe` and -- if plotting is enabled -- :meth:`~rics.performance.plot_run`. For full
    functionally these methods should be use directly.

    Args:
        candidate_method: A single method, collection of functions or a dict {label: function} of candidates.
        test_data: A single datum, or a dict ``{label: data}`` to evaluate candidates on.
        time_per_candidate: Desired runtime for each repetition per candidate label.
        plot: If ``True``, plot a figure using :meth:`~rics.performance.plot_run`.
        show: If ``True``, attempt to display the figure. Ignored when ``plot=False``.
        names: Level names for tuple keys in the data (creates new columns). See :func:`.plot_run` for details. Set to
            ``None`` to disable derived names when `test_data` is callable.
        progress: If ``True``, display a progress bar. Requires ``tqdm``.
        case_args: These are positional arguments for the `test_data` callable.
        kwargs: Shared keyword arguments for the `test_data` callable.
        **plot_kwargs: See :func:`.plot_run` for details. Ignored if ``plot=False``.

    Returns:
        A long-format DataFrame of results.

    Raises:
        ModuleNotFoundError: If Seaborn isn't installed and ``plot=True``.
        TypeError: If `args` or `kwargs` are set when `test_data` is not a callable.
        ValueError: If `args` is empty and `test_data` is a callable.

    See Also:
        The :func:`~rics.performance.plot_run` and :func:`~rics.performance.get_best` functions.
    """
    if plot:
        _verify_can_plot()

    timer: MultiCaseTimer[DataType, *Ts] = MultiCaseTimer(
        candidate_method,
        test_data,
        case_args=case_args,
        kwargs=kwargs,
    )
    run_results = timer.run(time_per_candidate=time_per_candidate, progress=progress)

    if names is None:
        names = ()
    else:
        names = [*names]
        if not names and timer.is_data_generated:
            names = timer.derive_names()

    data = to_dataframe(run_results, names=names)

    if plot:
        from matplotlib.pyplot import show as show_figure

        from ._plot import plot_run

        plot_run(data, names=names, **plot_kwargs)

        if show:
            show_figure(block=False)

    return data


def _verify_can_plot() -> None:
    from importlib.util import find_spec

    if find_spec("seaborn") is None:
        msg = (
            "Package `seaborn` not installed. Run of one:"
            "\n  pip install seaborn"
            "\n  pip install rics[plotting]"
            "\nto enable plots."
        )
        raise RuntimeError(msg)



from typing import Any

_SKIP_WARNING: bool = False


class ConfigureStuff:
    """Magical return type of :func:`configure_stuff`. The same caveats apply."""

    def __init__(self) -> None:
        self._configured: list[str] = []

    @property
    def configured(self) -> list[str]:
        """Things that were configured."""
        return self._configured

    def __repr__(self) -> str:
        return "👻 Configured some stuff just the way I like it!"


def configure_stuff(
    level: int | str = "INFO",
    rics_level: int | str = "INFO",
    id_translation_level: int | str = "INFO",
    matplotlib_level: int | str = "WARNING",
    logging: bool = True,
    pandas: bool = True,
    plotting: bool = True,
    **kwargs: Any,
) -> ConfigureStuff:
    """Configure a bunch of stuff to match my personal preferences. May do strange stuff 👻.

    .. warning::

       This function can and will change without warning, and will not be documented in the changelog. Don't use for
       anything important.

    Set :envvar:`JTWILI=true <JTWILI>` to disable the warning.

    Args:
        level: Log level for the root logger. Default is ``logging.INFO``.
        rics_level: Log level for the :mod:`rics` package. Default is ``logging.INFO``.
        id_translation_level: Log level for the :mod:`id_translation` package. Default is ``logging.INFO``.
        matplotlib_level: Log level for the :mod:`matplotlib` package. Default is ``logging.WARNING``.
        logging: If ``True``, attempt to perform logging configuration.
        pandas: If ``True``, attempt to perform pandas configuration.
        plotting: If ``True``, attempt to perform plotting (e.g. matplotlib) configuration.
        **kwargs: Keyword arguments for :func:`rics.logs.basic_config` and :py:func:`logging.basicConfig`.

    """
    import contextlib

    stuff = ConfigureStuff()
    configured = stuff.configured

    if logging:
        from .logs import basic_config

        basic_config(
            level=level,
            rics_level=rics_level,
            id_translation_level=id_translation_level,
            matplotlib_level=matplotlib_level,
            **kwargs,
        )
        configured.append("logging")

    if pandas:
        _configure_pandas()
        configured.append("pandas")

    if plotting:
        with contextlib.suppress(ModuleNotFoundError):
            from .plotting import configure

            configure()
            configured.append("plotting")

    _maybe_emit_warning()

    return stuff


def _configure_pandas() -> None:
    try:
        import pandas as pd
    except ModuleNotFoundError:
        return

    import contextlib

    pd.options.display.max_columns = 50
    pd.options.display.max_colwidth = 150
    pd.options.display.max_rows = 250
    pd.options.display.width = 0
    pd.options.display.float_format = "{:.6g}".format

    pd.options.mode.chained_assignment = "raise"
    if hasattr(pd.options.mode, "copy_on_write") and pd.options.mode.copy_on_write is None:
        pd.options.mode.copy_on_write = "warn"

    with contextlib.suppress(ImportError):
        pd.plotting.register_matplotlib_converters()


def _maybe_emit_warning() -> None:  # pragma: no cover
    import logging
    import os
    import sys
    import traceback
    from multiprocessing import process

    global _SKIP_WARNING  # noqa: PLW0603
    if _SKIP_WARNING:
        return
    _SKIP_WARNING = True

    if os.environ.get("JTWILI", "").lower() == "true":
        return

    if getattr(process.current_process(), "_inheriting", False):
        return  # From multiprocessing.spawn._check_not_importing_main

    try:
        get_ipython()  # type: ignore[name-defined]
        return  # noqa: TRY300 # Typically in a console or notebook
    except NameError:
        pass

    caller = traceback.format_stack()[-3]
    message = f"If you're seeing this in bad places, remove the call to rics.configure_stuff() in:\n{caller}"

    logger = logging.getLogger("rics")
    if logger.isEnabledFor(logging.WARNING):
        logger.warning(message)
    else:
        print(message, end="", file=sys.stderr)



import shutil

from rics.paths import AnyPath, any_path_to_path

from ._patch_notes import PatchNotes

BASE_LINES: list[str] = [
    "Changelog",
    "=========",
    "All notable changes to this project will be documented in this file.",
    "The format is based on `Keep a Changelog <https://keepachangelog.com/en/1.0.0/>`_",
    "and this project adheres to `Semantic Versioning <https://semver.org/spec/v2.0.0.html>`_.",
    "",
    ".. toctree::",
    "   :hidden:",
    "",
]


def write_changelog(
    output_dir: AnyPath,
    notes: list[PatchNotes],
    refs: dict[str, str],
) -> None:
    """Write changelog parts to several pages in an output directory.

    Args:
        output_dir: Output directory where pages will be placed. Only ``index.rst`` needs to be manually imported.
        notes: A list of patch notes, in the order in which they should be added to the TOC.
        refs: Title references.

    """
    root = any_path_to_path(output_dir)
    shutil.rmtree(root, ignore_errors=True)
    root.mkdir()

    index_lines = BASE_LINES.copy()

    for note in notes:
        if not note.sections:
            continue

        page = note.title.replace("[", "").replace("]", "").replace("#", "").lower().strip()
        index_lines.append(f"   {page}")
        file = root.joinpath(f"{page}.md")
        lines = ["# " + note.title + (f" ({note.date})" if note.date else "")]

        if note.extras:
            lines.extend(note.extras)
            lines.append("")

        for section, content in note.sections.items():
            lines.append("## " + section)
            lines.extend(content)

        lines.append("\n" + refs[note.title])
        file.write_text("\n".join(lines))

    root.joinpath("index.rst").write_text("\n".join(index_lines))



"""My personal little ML engineering library.

Environment variables
---------------------
.. envvar:: JTWILI

    Set ``JTWILI=true`` to disable the warning emitted by :func:`rics.configure_stuff`.


"""

import logging as _logging

from ._just_the_way_i_like_it import ConfigureStuff, configure_stuff

__all__ = [
    "ConfigureStuff",
    "__version__",
    "configure_stuff",
]

__version__ = "6.1.1"

_logging.getLogger(__name__).addHandler(_logging.NullHandler())



from typing import Any

from myst_parser.sphinx_ext.myst_refs import MystReferenceResolver
from sphinx.addnodes import pending_xref

_real = MystReferenceResolver.resolve_myst_ref_doc


def patch() -> None:
    if MystReferenceResolver.resolve_myst_ref_doc is _real:
        MystReferenceResolver.resolve_myst_ref_doc = _hack  # type: ignore[method-assign]
        print("patch: MystReferenceResolver.resolve_myst_ref_doc")


def _hack(self: MystReferenceResolver, node: pending_xref) -> Any:
    ref_docname: str = node["reftarget"]
    prefix, _, name = ref_docname.rpartition("/")

    if prefix and name in {"LICENSE", "CONTRIBUTING", "CODE_OF_CONDUCT"}:
        node["reftarget"] = name

    return _real(self, node)



import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal

import pandas as pd

from rics.performance.plot_types import Kind, Unit
from rics.performance.types import ResultsDict

from ._params import CatplotParams
from ._postprocessors import make_postprocessors

if TYPE_CHECKING:
    import seaborn


def plot_run(
    run_results: ResultsDict | pd.DataFrame,
    *,
    x: Literal["candidate", "data"] | None = None,
    horizontal: bool = False,
    unit: Unit | None = None,
    kind: Kind = "bar",
    names: Iterable[str] = (),
    **kwargs: Any,
) -> "seaborn.FacetGrid":
    """Create a :func:`seaborn.catplot` from run results.

    .. figure:: /_images/perf_plot_facets.png

       Comparison of best-per-group selection functions (from the
       :ref:`examples </documentation/examples/notebooks/performance/best-by-group/Best-by-Group.ipynb>`
       page).

    The `names` argument:
        Names may be passed in combination with ``row`` and/or ``col`` arguments to add facets to the
        :func:`seaborn.catplot`. If given, the keys in the test data must be of type ``tuple`` with the same length as
        `names`. For example, if your test data looks like this:

        .. code-block::

           test_data = {
               ("+", 2,  5): +(2**5),
               ("+", 9,  5): +(9**5),
               ("+", 10, 5): +(10**5),
               ("-", 10, 3): -(10**3),
               ("-", 5,  3): -(5**3),
           }

        you may pass

        .. code-block::

           plot_run(
               run_results = ...,
               names=["sign", "base", "exponent"],
               col="sign",
               row="exponent",
           )

        to plot each sign/exponent in a separate facet, comparing only the exponents in the subplots.

    Args:
        run_results: Output of :meth:`.MultiCaseTimer.run`.
        x: X-axis quantity; ``candidate'`` or ``'data'``. The other will be used as the hue.
        horizontal: If ``True``, plot timings on the X-axis instead. The `x` becomes the new Y-axis quantity.
        unit: Y-axis time :attr:`~rics.performance.plot_types.Unit`.
        kind: The :attr:`~rics.performance.plot_types.Kind` of plot to draw.
        names: Test data level names.
        **kwargs: Keyword arguments for :func:`seaborn.catplot`.

    Returns:
        A :class:`seaborn.FacetGrid`.

    Raises:
        ModuleNotFoundError: If Seaborn isn't installed.
        TypeError: For unknown `unit` arguments.
    """
    params = CatplotParams.make(run_results, x=x, horizontal=horizontal, unit=unit, kind=kind, names=names, **kwargs)
    return plot_params(params)


def plot_params(params: CatplotParams) -> "seaborn.axisgrid.FacetGrid":
    """Create a :func:`seaborn.catplot` from :class:`CatplotParams`."""
    from seaborn import catplot

    kwargs = params.to_kwargs()
    postprocessors = make_postprocessors(kwargs)
    facet_grid = catplot(**kwargs)

    logger = logging.getLogger(__package__).getChild("plot")
    if logger.isEnabledFor(logging.DEBUG):
        from rics.misc import get_public_module

        without_data = kwargs.copy()
        shape = without_data.pop("data").shape
        pretty_func = get_public_module(catplot) + "." + catplot.__qualname__
        logger.debug(
            f"Calling {pretty_func}(DataFrame[{' x '.join(map(str, shape))}], **kwargs) with:"
            f"\n - {params=}"
            f"\n - kwargs={without_data}"
            f"\n - {postprocessors=}",
            extra=without_data,
        )

    facet_grid.set_titles()

    for p in postprocessors:
        p(facet_grid)

    return facet_grid
